// cConstructModel.cpp

#include "cConstructModel.h"

#include <GL/gl.h>

#include <cassert>

using namespace std;

//-----------------------------------

void setupMatrix(float *mat, float *rot = NULL, float *pos = NULL)
{
  glPushMatrix();
  glLoadIdentity();

  if (pos != NULL) glTranslatef(pos[0], pos[1], pos[2]);

  if (rot != NULL) {
    glRotatef(rot[0], 1, 0, 0);
    glRotatef(rot[1], 0, 1, 0);
    glRotatef(rot[2], 0, 0, 1);
  };

  glGetFloatv(GL_MODELVIEW_MATRIX, mat);

  glPopMatrix();
}


void setupInvMatrix(float *mat, float *rot = NULL, float *pos = NULL)
{
  glPushMatrix();
  glLoadIdentity();

  if (rot != NULL) {
    glRotatef(-rot[2], 0, 0, 1);
    glRotatef(-rot[1], 0, 1, 0);
    glRotatef(-rot[0], 1, 0, 0);
  };
  if (pos != NULL) glTranslatef(-pos[0], -pos[1], -pos[2]);

  glGetFloatv(GL_MODELVIEW_MATRIX, mat);

  glPopMatrix();
}

//-----------------------------------


cConstruct::cConstruct()
{
  mName[0] = '\0';
  mVisible     = 1;
  mTransformed = false;
  mParent      = NULL;
  mVolume      = 0;
  mSurface     = 0;
  mDamage      = 0;
  //mPoseMatrixfv = NULL;
  mRotator3fv = NULL;
  mTranslator3fv = NULL;
}


void cConstruct::Calculate()
{
  mVolume      = 0;
  mSurface     = 0;
}


void cConstruct::ToWorld()
{
  if (mTransformed) return;
  else mTransformed = true;

  if (mParent) mParent->ToWorld();
  
  glPushAttrib(GL_TRANSFORM_BIT);
  glMatrixMode(GL_MODELVIEW);

  // Setup Inverse Matrix (Global-Coords to Local-Coords).
  glPushMatrix();
  glLoadIdentity();
  //if (mPoseInvMatrixfv != NULL) glMultMatrixf(mPoseInvMatrixfv);
  glTranslatef(-mPos.mV[X], -mPos.mV[Y], -mPos.mV[Z]);
  glRotatef(-mRot.mV[Z], 0, 0, 1);
  glRotatef(-mRot.mV[Y], 0, 1, 0);
  glRotatef(-mRot.mV[X], 1, 0, 0);
  if (mRotator3fv)
  {
    glRotatef(-mRotator3fv[Z], 0, 0, 1);
    glRotatef(-mRotator3fv[Y], 0, 1, 0);
    glRotatef(-mRotator3fv[X], 1, 0, 0);
  };
  if (mParent) glMultMatrixf(mParent->mInvMatrix.mM);
  glGetFloatv(GL_MODELVIEW_MATRIX, mInvMatrix.mM);
  glPopMatrix();


  // Setup Matrix (Local-Coords to Global-Coords).
  glPushMatrix();
  glLoadIdentity();
  if (mParent) glLoadMatrixf(mParent->mMatrix.mM);
  //if (mPoseMatrixfv != NULL) glMultMatrixf(mPoseMatrixfv);
  if (mRotator3fv)
  {
    glRotatef(mRotator3fv[X], 1, 0, 0);
    glRotatef(mRotator3fv[Y], 0, 1, 0);
    glRotatef(mRotator3fv[Z], 0, 0, 1);
  };
  glRotatef(mRot.mV[X], 1, 0, 0);
  glRotatef(mRot.mV[Y], 0, 1, 0);
  glRotatef(mRot.mV[Z], 0, 0, 1);
  glTranslatef(mPos.mV[X], mPos.mV[Y], mPos.mV[Z]);
  glGetFloatv(GL_MODELVIEW_MATRIX, mMatrix.mM);
  glPopMatrix();

  //mSphere.mV.Set(0, 0, 0);
  
  glPopAttrib();
}


void cConstruct::Draw()
{
  return;
  glPushMatrix();
  glMultMatrixf(mMatrix.mM);

  if (mVisible)
  {
  };

  glPopMatrix();
}


bool cConstruct::collideLine(float* a, float* b, float* point, float* normal, float* dist)
{
  cVector a_, b_;
  a_.Set(a[0], a[1], a[2]);
  b_.Set(b[0], b[1], b[2]);

  cMatrix im;
  memcpy(im.mM, mInvMatrix.mM, sizeof(float) * 16);
//  im.Apply(a_);
//  im.Apply(b_);

  cRay ray;
  ray.Set(a_, b_);

  cVector pnt;
  if (mSphere.Hit(ray, pnt))
  {
    cVector delta;
    delta.Set(b_, a_);
    delta.Normalise();
         if (fabs(delta.mV[0]) > 0.1) *dist = (pnt.mV[0] - a_.mV[0]) /  delta.mV[0];
    else if (fabs(delta.mV[1]) > 0.1) *dist = (pnt.mV[1] - a_.mV[1]) /  delta.mV[1];
    else if (fabs(delta.mV[2]) > 0.1) *dist = (pnt.mV[2] - a_.mV[2]) /  delta.mV[2];

    mDamage += 0.01f;
    if (dist >= 0)
    {
      normal[0] = -delta.mV[0];
      normal[1] = -delta.mV[1];
      normal[2] = -delta.mV[2];
      point[0] = pnt.mV[0];
      point[1] = pnt.mV[1];
      point[2] = pnt.mV[2];
      return true;
    };
  };

  return false;
}


//-----------------------------------


void cCriss::Calculate()
{
  mVolume = mWidth*mHeight*mLength+0.001;
  mSurface = 1.0*(mWidth*mHeight + mWidth*mLength + mHeight*mLength);
  mSphere.mR = sqrt(xsqr(mWidth/2.0) + xsqr(mHeight/2.0) + xsqr(mLength/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cCriss::Draw()
{
  float W2 = mWidth/2.0;
  float H2 = mHeight/2.0;
  float L2 = mLength/2.0;
  cVector p[8];

  glPushMatrix();
  glMultMatrixf(mMatrix.mM);

  if (mVisible) {

    p[0].Set(    -W2,     +H2,      0);
    p[1].Set(    -W2,     -H2,      0);
    p[2].Set(      0,     +H2,    -L2);
    p[3].Set(      0,     -H2,    -L2);

    p[4].Set(    +W2,     +H2,      0);
    p[5].Set(    +W2,     -H2,      0);
    p[6].Set(      0,     +H2,    +L2);
    p[7].Set(      0,     -H2,    +L2);

    float wm = 1.0-mDamage/mVolume;
    if (wm<0) wm = 0;
    if (wm>1) wm = 1;
    glColor4f(1.0, wm, wm, 1.0);
    glBegin(GL_QUADS);

    glTexCoord2f(0, 0);
    glVertex3fv(p[7].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[2].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[6].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[7].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[6].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[2].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[5].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[1].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[1].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[5].mV);

    glEnd();
  };

  glPopMatrix();
}


//-----------------------------------


void cCross::Calculate()
{
  mVolume = mWidth*mHeight*mLength+0.001;
  mSurface = 1.0*(mWidth*mHeight + mWidth*mLength + mHeight*mLength);
  mSphere.mR = sqrt(xsqr(mWidth/2.0) + xsqr(mHeight/2.0) + xsqr(mLength/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cCross::Draw()
{
  float W2 = mWidth/2.0;
  float H2 = mHeight/2.0;
  float L2 = mLength/2.0;
  cVector p[12];

  glPushMatrix();
  glMultMatrixf(mMatrix.mM);

  if (mVisible) {

    p[0].Set(    -W2,       0,    -L2);
    p[1].Set(    +W2,       0,    -L2);
    p[2].Set(      0,      H2,    -L2);
    p[3].Set(      0,     -H2,    -L2);

    p[4].Set(    -W2,       0,    +L2);
    p[5].Set(    +W2,       0,    +L2);
    p[6].Set(      0,      H2,    +L2);
    p[7].Set(      0,     -H2,    +L2);

    p[8].Set(    -W2,     -H2,      0);
    p[9].Set(    +W2,     -H2,      0);
    p[10].Set(   +W2,     +H2,      0);
    p[11].Set(   -W2,     +H2,      0);

    float wm = 1.0-mDamage/mVolume;
    if (wm<0) wm = 0;
    if (wm>1) wm = 1;
    glColor4f(1.0, wm, wm, 1.0);

    glBegin(GL_QUADS);

    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[1].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[5].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[4].mV);

    glTexCoord2f(1, 0);
    glVertex3fv(p[1].mV);
    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[5].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[2].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[6].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[7].mV);

    glTexCoord2f(1, 0);
    glVertex3fv(p[2].mV);
    glTexCoord2f(0, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[7].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[6].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[8].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[9].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[10].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[11].mV);

    glTexCoord2f(1, 0);
    glVertex3fv(p[9].mV);
    glTexCoord2f(0, 0);
    glVertex3fv(p[8].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[11].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[10].mV);

    glEnd();
  };

  glPopMatrix();
}


//-----------------------------------


void cQuader::Calculate()
{
  mVolume = mWidth*mHeight*mLength;
  mSurface = 2.0*(mWidth*mHeight + mWidth*mLength + mHeight*mLength);
  mSphere.mR = sqrt(xsqr(mWidth/2.0) + xsqr(mHeight/2.0) + xsqr(mLength/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cQuader::Draw()
{
  if (!mVisible) return;
  
  float wm = 1.0-mDamage/mVolume;
  if (wm<0) wm = 0;
  if (wm>1) wm = 1;
  glColor4f(1.0, wm, wm, 1.0);

  glPushMatrix();
    glMultMatrixf(mMatrix.mM);
        glScalef(mWidth*0.5f, mHeight*0.5f, mLength*0.5f);
            //glTranslatef(-0.5f, -0.5f, -0.5f);
            //glUnitBlock();
            glCenterUnitBlock();
  glPopMatrix();
}


//-----------------------------------


void cCylinder::Calculate()
{
  mVolume = M_PI * xsqr(mRadius) * mHeight;
  mSurface = 2.0 * M_PI * mRadius * (mRadius * mHeight);
  mSphere.mR = xmax(mRadius, (mHeight/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cCylinder::Draw()
{
  if (!mVisible) return;

  float wm = 1.0-mDamage/mVolume;
  if (wm<0) wm = 0;
  if (wm>1) wm = 1;
  glColor4f(1.0, wm, wm, 1.0);

  glPushMatrix();
    glMultMatrixf(mMatrix.mM);
        glScalef(mRadius, mHeight*0.5f, mRadius);
        glCenterUnitCylinder(5);
  glPopMatrix();
}


//-----------------------------------


#include <map>


cConstructModel::cConstructModel(char* descriptor)
{
    if (createContext(descriptor)) {
        printf("cConstructModel(%s): creation error\n", descriptor);
    }
}


bool cConstructModel::createContext(char* descriptor)
{
  this->rotator3fv = NULL;
  this->translator3fv = NULL;
  setupMatrix(this->matrix);
  setupInvMatrix(this->invmatrix);
  
  if (descriptor == NULL) return true;

  FILE* f = fopen(descriptor, "rt");
  // If Model not available then return NULL.
  if (f == NULL) return true;
  
  if (this->loadSelf(f)) {
    assert("Couldn't load construct" == NULL);
    return true;
  };
  
  fclose(f);

  // Initialize hooks for individual construct-blocks.
  for (int i = 0; i < this->constructs.size(); i++)
  {
    this->rotators.push_back(NULL);
    this->translators.push_back(NULL);
    this->matrices.push_back(new float[16]);
    this->invmatrices.push_back(new float[16]);
  };
  
  return false;
}


bool cConstructModel::loadSelf(FILE *f)
{
  int           nparts, s;
  cCylinder    *cy;
  cQuader      *q;
  cCriss       *ci;
  cCross       *cr;
  float        x, y, z, a, b, c;

  typedef map<string,int> symmap;
  symmap conmap;
  conmap.insert(symmap::value_type("construct", cConstruct::Construct));
  conmap.insert(symmap::value_type("cylinder", cConstruct::Cylinder));
  conmap.insert(symmap::value_type("quader", cConstruct::Quader));
  conmap.insert(symmap::value_type("criss", cConstruct::Criss));
  conmap.insert(symmap::value_type("cross", cConstruct::Cross));

  char line[256];
  char par[64], nam[64], con[64];

  fscanf(f, "%i %i\n", &nparts, &s);

  cConstruct* construct;
  for (int k=0; k<nparts; k++)
  {
    fgets(line, 256, f);

    string contype_s;

    sscanf(line, "%s %s %s   %f %f %f", par, nam, con, &x, &y, &z);

    switch (conmap[con]) {

      case cConstruct::Construct: {
        construct = new cConstruct;
        strcpy(construct->mParentName, par);
        strcpy(construct->mName, nam);
        construct->mType = cConstruct::Construct;
        construct->mPos.Set(x, y, z);
      } break;
      case cConstruct::Cylinder: {
        sscanf(line, "%s %s %s   %f %f %f   %f %f", par, nam, con,  &x, &y, &z,  &a, &b);
        construct = new cCylinder;
        strcpy(construct->mParentName, par);
        strcpy(construct->mName, nam);
        construct->mType = cConstruct::Cylinder;
        construct->mPos.Set(x, y, z);
        cy = (cCylinder*) construct;
        cy->mRadius = a;
        cy->mHeight = b;
        construct->Calculate();
      } break;
      case cConstruct::Quader: {
        sscanf(line, "%s %s %s   %f %f %f   %f %f %f", par, nam, con,  &x, &y, &z,  &a, &b, &c);
        construct = new cQuader;
        strcpy(construct->mParentName, par);
        strcpy(construct->mName, nam);
        construct->mType = cConstruct::Quader;
        construct->mPos.Set(x, y, z);
        q = (cQuader*) construct;
        q->mWidth = a;
        q->mHeight = b;
        q->mLength = c;
        construct->Calculate();
      } break;
      case cConstruct::Criss: {
        sscanf(line, "%s %s %s   %f %f %f   %f %f %f", par, nam, con,  &x, &y, &z,  &a, &b, &c);
        construct = new cCriss;
        strcpy(construct->mParentName, par);
        strcpy(construct->mName, nam);
        construct->mType = cConstruct::Criss;
        construct->mPos.Set(x, y, z);
        ci = (cCriss*) construct;
        ci->mWidth = a;
        ci->mHeight = b;
        ci->mLength = c;
        construct->Calculate();
      } break;
      case cConstruct::Cross: {
        sscanf(line, "%s %s %s   %f %f %f   %f %f %f", par, nam, con,  &x, &y, &z,  &a, &b, &c);
        construct = new cCross;
        strcpy(construct->mParentName, par);
        strcpy(construct->mName, nam);
        construct->mType = cConstruct::Cross;
        construct->mPos.Set(x, y, z);
        cr = (cCross*) construct;
        cr->mWidth = a;
        cr->mHeight = b;
        cr->mLength = c;
        construct->Calculate();
      } break;
      default: assert("No such construct" == NULL); break;
    };
    constructs.push_back(construct);
  };

  cConstruct *construct_i;
  cConstruct *construct_j;
  list<cConstruct*>::iterator i;
  list<cConstruct*>::iterator j;

  for (i = constructs.begin(); i != constructs.end();)
  {
    construct_i = *i; i++;
    construct_i->mParent = NULL; //construct;
    if (strcmp(construct_i->mParentName, "NULL") == 0) continue;
    for (j = constructs.begin(); j != constructs.end();)
    {
      construct_j = *j; j++;
      if (construct_i == construct_j) continue;
      if (strcmp(construct_i->mParentName, construct_j->mName) == 0)
      {
        construct_i->mParent = construct_j;
        break;
      };
    };
  };

  return false;
}


void cConstructModel::transformSelf()
{
  setupInvMatrix(this->invmatrix, this->rotator3fv, this->translator3fv);
  setupMatrix(this->matrix, this->rotator3fv, this->translator3fv);

  cConstruct *construct;
  list<cConstruct*>::iterator i;
  float* rot3fv;
  float* tra3fv;
  list<float*>::iterator r;
  list<float*>::iterator t;
  for (i = this->constructs.begin(), r = this->rotators.begin(), t = this->translators.begin(); i != this->constructs.end();)
  {
    rot3fv = *r; r++;
    tra3fv = *t; t++;
    construct = *i; i++;
    construct->mRotator3fv = rot3fv;
    construct->mTranslator3fv = tra3fv;
    construct->mTransformed = false;
  };

  list<float*>::iterator j;
  list<float*>::iterator k;
  float* mat;
  for (i = this->constructs.begin(), j = this->matrices.begin(), k = this->invmatrices.begin(); i != this->constructs.end();)
  {
    construct = *i; i++;
    construct->ToWorld();

    mat = *j; j++;
    memcpy(mat, construct->mMatrix.mM, sizeof(float) * 16);

    mat = *k; k++;
    memcpy(mat, construct->mInvMatrix.mM, sizeof(float) * 16);
  };
}


void cConstructModel::drawSelf()
{
  glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
  glPushMatrix();

  // Put model to required position.
  glMultMatrixf(this->matrix);

  /*
  glDisable(GL_TEXTURE_2D);
  glColor4f(0.6, 0.6, 0.6, 0.4);
  glutSolidSphere(1.4f, 3, 3);
  glEnable(GL_TEXTURE_2D);
  */

  //glDisable(GL_TEXTURE_2D);
  glEnable(GL_TEXTURE_2D);
  cConstruct *construct;
  list<cConstruct*>::iterator i;
  list<float*>::iterator j;
  float* mat;
  for (i = this->constructs.begin(), j = this->matrices.begin(); i != this->constructs.end();)
  {
    construct = *i; i++;
    mat = *j; j++;
    memcpy(construct->mMatrix.mM, mat, sizeof(float) * 16);
    construct->Draw();
  };

  glPopMatrix();
  glPopAttrib();
}


cConstruct* cConstructModel::findConstruct(char* name, list<float*>::iterator &it)
{
  cConstruct *construct;
  list<cConstruct*>::iterator i;
  list<float*>::iterator j;
  for (i = this->constructs.begin(); i != this->constructs.end();)
  {
    construct = *i; i++;
    if (strcmp(construct->mName, name) == 0) return construct;
    it++;
  };
  return NULL;
}


void cConstructModel::setRotator(float* rot3fv)
{
  this->rotator3fv = rot3fv;
}


void cConstructModel::setTranslator(float* tra3fv)
{
  this->translator3fv = tra3fv;
}


void cConstructModel::setConstructRotator(char* name, float* rot3fv)
{
  list<float*>::iterator i = this->rotators.begin();
  if (findConstruct(name, i))
  {
    *i = rot3fv;
  };
}


void cConstructModel::setConstructTranslator(char* name, float* tra3fv)
{
  list<float*>::iterator i = this->translators.begin();
  if (findConstruct(name, i))
  {
    *i = tra3fv;
  };
}


float* cConstructModel::getMatrix()
{
  return this->matrix;
}


float* cConstructModel::getInvMatrix()
{
  return this->invmatrix;
}


float* cConstructModel::getConstructMatrix(char* name)
{
  list<float*>::iterator i = this->matrices.begin();
  if (findConstruct(name, i))
  {
    float* matfv = *i;
    return matfv;
  };
  return NULL;
}


float* cConstructModel::getConstructInvMatrix(char* name)
{
  list<float*>::iterator i = this->invmatrices.begin();
  if (findConstruct(name, i))
  {
    float* matfv = *i;
    return matfv;
  };
  return NULL;
}


void cConstructModel::collideLine(float* a, float* b, rImpact *&result)
{
  // Transform line into local coordinate system
  // by the inverse matrix.
  cMatrix m, n;
  memcpy(m.mM, this->invmatrix, sizeof(float) * 16);
  cVector a_, b_;

  result = NULL;

  list<cConstruct*>::iterator i;
  list<float*>::iterator j;
  for (i = this->constructs.begin(), j = this->invmatrices.begin(); i != this->constructs.end(); i++, j++)
  {
    cConstruct* construct = *i;
    float* mat = *j;
    
    // Transform the line again into the more local
    // coordinate system of the individual construct.
    memcpy(construct->mInvMatrix.mM, mat, sizeof(float) * 16);
    memcpy(n.mM, mat, sizeof(float) * 16);

    cMatrix o;
    o = m;
    o.Mul(n);

    a_.Set(a[0], a[1], a[2]);
    b_.Set(b[0], b[1], b[2]);

    o.Apply(a_);
    o.Apply(b_);

    float normal[3], point[3], dist;
    if ( construct->collideLine(a_.mV, b_.mV, normal, point, &dist) ) {
        //printf("#1 collideLine dist = %.6f\n", dist);
        if (dist >= 0.0f && dist <= 1.0f) {
            //printf("#2 collideLine dist = %.2f\n", dist);
            result = new rImpact();
            result->intsec_distance = dist;
            memcpy(result->intsec_point, point, sizeof(float)*3);
            memcpy(result->intsec_normal, normal, sizeof(float)*3);
            break;
        }
    }
    
  } // for constructs/inverse-matrices
}



